__version__ = '0.4.0'
from .causalEGM import CausalEGM, VariationalCausalEGM
from .util import Base_sampler, Semi_acic_sampler, Sim_Hirano_Imbens_sampler, Sim_Sun_sampler, Sim_Colangelo_sampler, Semi_Twins_sampler


from .causalEGM import CausalEGM
from .util import *
import argparse
from CausalEGM import __version__


def main(args=None):

    parser = argparse.ArgumentParser('causalEGM',
                                     description=f'CausalEGM: A general causal inference framework by encoding generative modeling - v{__version__}')
    parser.add_argument('-output_dir', dest='output_dir', type=str,
                        help="Output directory", required=True)
    parser.add_argument('-input', dest='input', type=str,
                        help="Input data file must be in csv or txt or npz format", required=True)
    parser.add_argument('-dataset', dest='dataset', type=str,default='Mydata',
                        help="Dataset name")
    parser.add_argument('--save-model', default=True, action=argparse.BooleanOptionalAction,
                        help="whether to save model.")
    parser.add_argument('--binary-treatment', default=True, action=argparse.BooleanOptionalAction,
                        help="whether use binary treatment setting.")

    #model hypterparameters
    parser.add_argument('-z_dims', dest='z_dims', type=int, nargs='+', default=[3,6,3,6],
                        help='Latent dimensions of the four encoder outputs e(V)_0~3.')
    parser.add_argument('-lr', dest='lr', type=float, default=0.0002,
                        help="Learning rate for the optimizer (default: 0.0002).")
    parser.add_argument('-alpha', dest='alpha', type=float, default=1.,
                        help="Coefficient for reconstruction loss (default: 1).")
    parser.add_argument('-beta', dest='beta', type=float, default=1.,
                        help="Coefficient for treatment and outcome MSE loss (default: 1).")
    parser.add_argument('-gamma', dest='gamma', type=float, default=10.,
                        help="Coefficient for gradient penalty loss (default: 10).")
    parser.add_argument('-g_d_freq', dest='g_d_freq', type=int, default=5,
                        help="Frequency for updating discriminators and generators (default: 5).")
    #network hyperparameters
    parser.add_argument('-g_units', dest='g_units', type=int, nargs='+', default=[64,64,64,64,64],
                        help='Number of units for generator/decoder network (default: [64,64,64,64,64]).')
    parser.add_argument('-e_units', dest='e_units', type=int, nargs='+', default=[64,64,64,64,64],
                        help='Number of units for encoder network (default: [64,64,64,64,64]).')
    parser.add_argument('-f_units', dest='f_units', type=int, nargs='+', default=[64,32,8],
                        help='Number of units for f network (default: [64,32,8]).')
    parser.add_argument('-h_units', dest='h_units', type=int, nargs='+', default=[64,32,8],
                        help='Number of units for h network (default: [64,32,8]).')
    parser.add_argument('-dz_units', dest='dz_units', type=int, nargs='+', default=[64,32,8],
                        help='Number of units for discriminator network in latent space (default: [64,32,8]).')
    parser.add_argument('-dv_units', dest='dv_units', type=int, nargs='+', default=[64,32,8],
                        help='Number of units for discriminator network in confounder space (default: [64,32,8]).')
    parser.add_argument('--use-z-rec', default=True, action=argparse.BooleanOptionalAction,
                        help="Use the reconstruction for latent features.")
    parser.add_argument('--use-v-gan', default=True, action=argparse.BooleanOptionalAction,
                        help="Use the GAN distribution match for covariates.")
    #training parameters
    parser.add_argument('-batch_size', dest='batch_size', type=int,
                        default=32, help='Batch size (default: 32).')
    parser.add_argument('-n_iter', dest='n_iter', type=int, default=30000,
                        help="Number of iterations (default: 30000).")
    parser.add_argument('-startoff', dest='startoff', type=int, default=0,
                        help="Iteration for starting evaluation (default: 0).")
    parser.add_argument('-batches_per_eval', dest='batches_per_eval', type=int, default=500,
                        help="Number of iterations per evaluation (default: 500).")
    parser.add_argument('-save_format', dest='save_format', type=str,default='txt',
                        help="Saving format (default: txt)")
    parser.add_argument('--save_res', default=True, action=argparse.BooleanOptionalAction,
                        help="Whether to save results during training.")
    #Random seed control
    parser.add_argument('-seed', dest='seed', type=int, default=123,
                        help="Random seed for reproduction (default: 123).")
    args = parser.parse_args()
    params = vars(args)
    data = parse_file(args.input)
    params['v_dim'] = data[-1].shape[1]
    model = CausalEGM(params,random_seed=args.seed)
    print('Start training...')
    model.train(data=data, batch_size=args.batch_size, n_iter=args.n_iter, batches_per_eval=args.batches_per_eval, 
                startoff=args.startoff, save_format=args.save_format)

if __name__ == "__main__":
    main()



import tensorflow as tf

class BaseFullyConnectedNet(tf.keras.Model):
    """ Generator network.
    """
    def __init__(self, input_dim, output_dim, model_name, nb_units=[256, 256, 256], batchnorm=False):  
        super(BaseFullyConnectedNet, self).__init__()
        self.input_layer = tf.keras.layers.Input((input_dim,))
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.model_name = model_name
        self.nb_units = nb_units
        self.batchnorm = batchnorm
        self.all_layers = []
        """ Builds the FC stacks. """
        for i in range(len(nb_units) + 1):
            units = self.output_dim if i == len(nb_units) else self.nb_units[i]
            fc_layer = tf.keras.layers.Dense(
                units = units,
                activation = None,
                kernel_regularizer = tf.keras.regularizers.L2(2.5e-5)
            )   
            norm_layer = tf.keras.layers.BatchNormalization()
            self.all_layers.append([fc_layer, norm_layer])
        
        self.out = self.call(self.input_layer)

    def call(self, inputs, training=True):
        """ Return the output of the Generator.
        Args:
            inputs: tensor with shape [batch_size, input_dim]
        Returns:
            Output of Generator.
            float32 tensor with shape [batch_size, output_dim]
        """
        for i, layers in enumerate(self.all_layers[:-1]):
            # Run inputs through the sublayers.
            fc_layer, norm_layer = layers
            with tf.name_scope("%s_layer_%d" % (self.model_name, i+1)):
                x = fc_layer(inputs) if i==0 else fc_layer(x)
                if self.batchnorm:
                    x = norm_layer(x)
                x = tf.keras.layers.LeakyReLU(alpha=0.2)(x)
        fc_layer, norm_layer = self.all_layers[-1]
        with tf.name_scope("%s_layer_ouput" % self.model_name):
            output = fc_layer(x)
            # No activation func at last layer
            #x = tf.keras.layers.LeakyReLU(alpha=0.2)(x)
        return output

class Discriminator(tf.keras.Model):
    """Discriminator network.
    """
    def __init__(self, input_dim, model_name, nb_units=[256, 256], batchnorm=True):  
        super(Discriminator, self).__init__()
        self.input_layer = tf.keras.layers.Input((input_dim,))
        self.input_dim = input_dim
        self.model_name = model_name
        self.nb_units = nb_units
        self.batchnorm = batchnorm
        self.all_layers = []
        """Builds the FC stacks."""
        for i in range(len(self.nb_units)+1):
            units = 1 if i == len(self.nb_units) else self.nb_units[i]
            fc_layer = tf.keras.layers.Dense(
                units = units,
                activation = None
            )
            norm_layer = tf.keras.layers.BatchNormalization()

            self.all_layers.append([fc_layer, norm_layer])
        self.out = self.call(self.input_layer)

    def call(self, inputs, training=True):
        """Return the output of the Discriminator network.
        Args:
            inputs: tensor with shape [batch_size, input_dim]
        Returns:
            Output of Discriminator.
            float32 tensor with shape [batch_size, 1]
        """
            
        for i, layers in enumerate(self.all_layers[:-1]):
            # Run inputs through the sublayers.
            fc_layer, norm_layer = layers
            with tf.name_scope("%s_layer_%d" % (self.model_name,i+1)):
                x = fc_layer(inputs) if i==0 else fc_layer(x)
                if self.batchnorm:
                    x = norm_layer(x)
                x = tf.keras.activations.tanh(x)
                #x = tf.keras.layers.LeakyReLU(alpha=0.2)(x)
        fc_layer, norm_layer = self.all_layers[-1]
        with tf.name_scope("%s_layer_ouput" % self.model_name):
            output = fc_layer(x)
        return output


import tensorflow as tf
from .model import BaseFullyConnectedNet, Discriminator
import numpy as np
from .util import *
import dateutil.tz
import datetime
import os, sys

class CausalEGM(object):
    """Implementation of the CausalEGM model.

    Parameters
    ----------
    params
        Dict object denoting the hyperparameters for deployments and building the model architecture. 
        See examples under the ``src/configs`` folder.
    timestamp
        Str object denoting the timestemp for specificing when the model is instanced. Default: ``None``.
    random_seed
        Int object denoting the random seed for controling randomness. Default: ``None``.

    Examples
    --------
    >>> from CausalEGM import CausalEGM, Sim_Hirano_Imbens_sampler
    >>> import yaml
    >>> params = yaml.safe_load(open('src/configs/Sim_Hirano_Imbens.yaml', 'r'))
    >>> x,y,v = Sim_Hirano_Imbens_sampler(batch_size=32).load_all()
    >>> model = CausalEGM(params=params,random_seed=12)
    >>> model.train(data=[x,y,v],n_iter=30000,save_format='npy')
    """
    def __init__(self, params, timestamp=None, random_seed=None):
        super(CausalEGM, self).__init__()
        self.params = params
        self.timestamp = timestamp
        if random_seed is not None:
            tf.keras.utils.set_random_seed(random_seed)
            os.environ['TF_DETERMINISTIC_OPS'] = '1'
        self.g_net = BaseFullyConnectedNet(input_dim=sum(params['z_dims']),output_dim = params['v_dim'], 
                                        model_name='g_net', nb_units=params['g_units'])
        self.e_net = BaseFullyConnectedNet(input_dim=params['v_dim'],output_dim = sum(params['z_dims']), 
                                        model_name='e_net', nb_units=params['e_units'])
        self.dz_net = Discriminator(input_dim=sum(params['z_dims']),model_name='dz_net',
                                        nb_units=params['dz_units'])
        self.dv_net = Discriminator(input_dim=params['v_dim'],model_name='dv_net',
                                        nb_units=params['dv_units'])

        self.f_net = BaseFullyConnectedNet(input_dim=1+params['z_dims'][0]+params['z_dims'][1],
                                        output_dim = 1, model_name='f_net', nb_units=params['f_units'])
        self.h_net = BaseFullyConnectedNet(input_dim=params['z_dims'][0]+params['z_dims'][2],
                                        output_dim = 1, model_name='h_net', nb_units=params['h_units'])

        self.g_e_optimizer = tf.keras.optimizers.Adam(params['lr'], beta_1=0.5, beta_2=0.9)
        self.d_optimizer = tf.keras.optimizers.Adam(params['lr'], beta_1=0.5, beta_2=0.9)
        self.z_sampler = Gaussian_sampler(mean=np.zeros(sum(params['z_dims'])), sd=1.0)

        self.initialize_nets()
        if self.timestamp is None:
            now = datetime.datetime.now(dateutil.tz.tzlocal())
            self.timestamp = now.strftime('%Y%m%d_%H%M%S')
        
        self.checkpoint_path = "{}/checkpoints/{}/{}".format(
            params['output_dir'], params['dataset'], self.timestamp)

        if self.params['save_model'] and not os.path.exists(self.checkpoint_path):
            os.makedirs(self.checkpoint_path)
        
        self.save_dir = "{}/results/{}/{}".format(
            params['output_dir'], params['dataset'], self.timestamp)

        if self.params['save_res'] and not os.path.exists(self.save_dir):
            os.makedirs(self.save_dir)   

        self.ckpt = tf.train.Checkpoint(g_net = self.g_net,
                                   e_net = self.e_net,
                                   dz_net = self.dz_net,
                                   dv_net = self.dv_net,
                                   f_net = self.f_net,
                                   h_net = self.h_net,
                                   g_e_optimizer = self.g_e_optimizer,
                                   d_optimizer = self.d_optimizer)
        self.ckpt_manager = tf.train.CheckpointManager(self.ckpt, self.checkpoint_path, max_to_keep=3)                 

        if self.ckpt_manager.latest_checkpoint:
            self.ckpt.restore(self.ckpt_manager.latest_checkpoint)
            print ('Latest checkpoint restored!!') 
        
    def get_config(self):
        """Get the parameters CausalEGM model."""

        return {
                "params": self.params,
        }
    
    def initialize_nets(self, print_summary = False):
        """Initialize all the networks in CausalEGM."""

        self.g_net(np.zeros((1, sum(self.params['z_dims']))))
        self.e_net(np.zeros((1, self.params['v_dim'])))
        self.dz_net(np.zeros((1, sum(self.params['z_dims']))))
        self.dv_net(np.zeros((1, self.params['v_dim'])))
        self.f_net(np.zeros((1, 1+self.params['z_dims'][0]+self.params['z_dims'][1])))
        self.h_net(np.zeros((1, self.params['z_dims'][0]+self.params['z_dims'][2])))
        if print_summary:
            print(self.g_net.summary())
            print(self.h_net.summary())
            print(self.dz_net.summary())
            print(self.f_net.summary())    
            print(self.h_net.summary()) 

    @tf.function
    def train_gen_step(self, data_z, data_v, data_x, data_y):
        """Training step for the generators in the CausalEGM model.

        Parameters
        ----------
        data_z
            Numpy.ndarray denoting latent features with shape [batch_size, z_dim]. 
        data_v
            Numpy.ndarray denoting covariants with shape [batch_size, v_dim]. 
        data_x
            Numpy.ndarray denoting treatment data with shape [batch_size, 1]. 
        data_y
            Numpy.ndarray denoting outcome data with shape [batch_size, 1]. 

        Returns
        --------
        g_loss_adv
            Float denoting G generator loss.
        e_loss_adv
            Float denoting E generator loss.
        l2_loss_v
            Float denoting V reconstruction loss.
        l2_loss_z
            Float denoting Z reconstruction loss.
        l2_loss_x
            Float denoting treatment reconstruction loss.
        l2_loss_y
            Float denoting outcome reconstruction loss.   
        g_e_loss
            Float denoting G E combined loss.      
        """ 
        with tf.GradientTape(persistent=True) as gen_tape:
            #data_x = tf.cast(data_x, tf.float32)
            data_v_ = self.g_net(data_z)
            data_z_ = self.e_net(data_v)

            data_z0 = data_z_[:,:self.params['z_dims'][0]]
            data_z1 = data_z_[:,self.params['z_dims'][0]:sum(self.params['z_dims'][:2])]
            data_z2 = data_z_[:,sum(self.params['z_dims'][:2]):sum(self.params['z_dims'][:3])]
            data_z3 = data_z_[:-self.params['z_dims'][3]:]

            data_z__= self.e_net(data_v_)
            data_v__ = self.g_net(data_z_)
            
            data_dv_ = self.dv_net(data_v_)
            data_dz_ = self.dz_net(data_z_)
            
            l2_loss_v = tf.reduce_mean((data_v - data_v__)**2)
            l2_loss_z = tf.reduce_mean((data_z - data_z__)**2)
            
            g_loss_adv = -tf.reduce_mean(data_dv_)
            e_loss_adv = -tf.reduce_mean(data_dz_)

            data_y_ = self.f_net(tf.concat([data_z0, data_z1, data_x], axis=-1))
            data_x_ = self.h_net(tf.concat([data_z0, data_z2], axis=-1))
            if self.params['binary_treatment']:
                data_x_ = tf.sigmoid(data_x_)
            l2_loss_x = tf.reduce_mean((data_x_ - data_x)**2)
            l2_loss_y = tf.reduce_mean((data_y_ - data_y)**2)
            g_e_loss = self.params['use_v_gan']*g_loss_adv+e_loss_adv+self.params['alpha']*(l2_loss_v + self.params['use_z_rec']*l2_loss_z) \
                        + self.params['beta']*(l2_loss_x+l2_loss_y)

        # Calculate the gradients for generators and discriminators
        g_e_gradients = gen_tape.gradient(g_e_loss, self.g_net.trainable_variables+self.e_net.trainable_variables+\
                                        self.f_net.trainable_variables+self.h_net.trainable_variables)
        
        # Apply the gradients to the optimizer
        self.g_e_optimizer.apply_gradients(zip(g_e_gradients, self.g_net.trainable_variables+self.e_net.trainable_variables+\
                                            self.f_net.trainable_variables+self.h_net.trainable_variables))
        return g_loss_adv, e_loss_adv, l2_loss_v, l2_loss_z, l2_loss_x, l2_loss_y, g_e_loss

    @tf.function
    def train_disc_step(self, data_z, data_v):
        """Training step for the discrinimator(s) in the CausalEGM model.

        Parameters
        ----------
        data_z
            Numpy.ndarray denoting latent features with shape [batch_size, z_dim]. 
        data_v
            Numpy.ndarray denoting covariants with shape [batch_size, v_dim]. 

        Returns
        --------
        dv_loss
            Float denoting V discrinimator loss.
        dz_loss
            Float denoting Z discrinimator loss.
        d_loss
            Float denoting combined discrinimator(s) loss.
        """ 
        epsilon_z = tf.random.uniform([],minval=0., maxval=1.)
        epsilon_v = tf.random.uniform([],minval=0., maxval=1.)
        with tf.GradientTape(persistent=True) as disc_tape:
            data_v_ = self.g_net(data_z)
            data_z_ = self.e_net(data_v)
            data_z_hat = data_z*epsilon_z + data_z_*(1-epsilon_z)
            data_v_hat = data_v*epsilon_v + data_v_*(1-epsilon_v)
            
            with tf.GradientTape() as gp_tape_z:
                gp_tape_z.watch(data_z_hat)
                data_dz_hat = self.dz_net(data_z_hat)
            with tf.GradientTape() as gp_tape_v:
                gp_tape_v.watch(data_v_hat)
                data_dv_hat = self.dv_net(data_v_hat)
                
            data_dv_ = self.dv_net(data_v_)
            data_dz_ = self.dz_net(data_z_)
            
            data_dv = self.dv_net(data_v)
            data_dz = self.dz_net(data_z)
            
            dz_loss = -tf.reduce_mean(data_dz) + tf.reduce_mean(data_dz_)
            dv_loss = -tf.reduce_mean(data_dv) + tf.reduce_mean(data_dv_)
            
            #gradient penalty for z
            grad_z = gp_tape_z.gradient(data_dz_hat, data_z_hat) #(bs,z_dim)
            grad_norm_z = tf.sqrt(tf.reduce_sum(tf.square(grad_z), axis=1))#(bs,) 
            gpz_loss = tf.reduce_mean(tf.square(grad_norm_z - 1.0))
            
            #gradient penalty for v
            grad_v = gp_tape_v.gradient(data_dv_hat, data_v_hat) #(bs,v_dim)
            grad_norm_v = tf.sqrt(tf.reduce_sum(tf.square(grad_v), axis=1))#(bs,) 
            gpv_loss = tf.reduce_mean(tf.square(grad_norm_v - 1.0))
            
            d_loss = self.params['use_v_gan']*dv_loss + dz_loss + \
                    self.params['gamma']*(gpz_loss + self.params['use_v_gan']*gpv_loss)

        # Calculate the gradients for generators and discriminators
        d_gradients = disc_tape.gradient(d_loss, self.dz_net.trainable_variables+self.dv_net.trainable_variables)
        
        # Apply the gradients to the optimizer
        self.d_optimizer.apply_gradients(zip(d_gradients, self.dz_net.trainable_variables+self.dv_net.trainable_variables))
        return dv_loss, dz_loss, d_loss

    def train(self, data=None, data_file=None, sep='\t', header=0, normalize=False,
            batch_size=32, n_iter=30000, batches_per_eval=500, batches_per_save=10000,
            startoff=0, verbose=1, save_format='txt'):
        """
        Train a CausalEGM model given the input data.
        
        Parameters
        ----------
        data
            List object containing the triplet data [X,Y,V]. Default: ``None``.
        data_file
            Str object denoting the path to the input file (csv, txt, npz).
        sep
            Str object denoting the delimiter for the input file. Default: ``\t``.
        header
            Int object denoting row number(s) to use as the column names. Default: ``0``.
        normalize
            Bool object denoting whether apply standard normalization to covariates. Default: ``False``.
        batch_size
            Int object denoting the batch size in training. Default: ``32``.
        n_iter
            Int object denoting the training iterations. Default: ``30000``.
        batches_per_eval
            Int object denoting the number of iterations per evaluation. Default: ``500``.
        batches_per_save
            Int object denoting the number of iterations per save. Default: ``10000``.
        startoff
            Int object denoting the beginning iterations to jump without save and evaluation. Defalt: ``0``.
        verbose
            Bool object denoting whether showing the progress bar. Default: ``False``.
        save_format
            Str object denoting the format (csv, txt, npz) to save the results. Default: ``txt``.
        """
        
        if self.params['save_res']:
            f_params = open('{}/params.txt'.format(self.save_dir),'w')
            f_params.write(str(self.params))
            f_params.close()
        if data is None and data_file is None:
            self.data_sampler = Dataset_selector(self.params['dataset'])(batch_size=batch_size)
        elif data is not None:
            if len(data) != 3:
                print('Data imcomplete error, please provide pair-wise (X, Y, V) in a list or tuple.')
                sys.exit()
            else:
                self.data_sampler = Base_sampler(x=data[0],y=data[1],v=data[2],batch_size=batch_size,normalize=normalize)
        else:
            data = parse_file(data_file, sep, header, normalize)
            self.data_sampler = Base_sampler(x=data[0],y=data[1],v=data[2],batch_size=batch_size,normalize=normalize)

        best_loss = np.inf
        for batch_idx in range(n_iter+1):
            for _ in range(self.params['g_d_freq']):
                batch_x, batch_y, batch_v = self.data_sampler.next_batch()
                batch_z = self.z_sampler.get_batch(len(batch_x))
                dv_loss, dz_loss, d_loss = self.train_disc_step(batch_z, batch_v)

            batch_x, batch_y, batch_v = self.data_sampler.next_batch()
            batch_z = self.z_sampler.get_batch(len(batch_x))
            g_loss_adv, e_loss_adv, l2_loss_v, l2_loss_z, l2_loss_x, l2_loss_y, g_e_loss = self.train_gen_step(batch_z, batch_v, batch_x, batch_y)

            if batch_idx % batches_per_eval == 0:
                loss_contents = '''Iteration [%d] : g_loss_adv [%.4f], e_loss_adv [%.4f],\
                l2_loss_v [%.4f], l2_loss_z [%.4f], l2_loss_x [%.4f],\
                l2_loss_y [%.4f], g_e_loss [%.4f], dv_loss [%.4f], dz_loss [%.4f], d_loss [%.4f]''' \
                %(batch_idx, g_loss_adv, e_loss_adv, l2_loss_v, l2_loss_z, l2_loss_x, l2_loss_y, g_e_loss,
                dv_loss, dz_loss, d_loss)
                if verbose:
                    print(loss_contents)
                causal_pre, mse_x, mse_y = self.evaluate(self.data_sampler.load_all())
                if batch_idx >= startoff and mse_y < best_loss:
                    best_loss = mse_y
                    self.best_causal_pre = causal_pre
                    self.best_batch_idx = batch_idx
                    if self.params['save_model']:
                        ckpt_save_path = self.ckpt_manager.save(batch_idx)
                        #print('Saving checkpoint for iteration {} at {}'.format(batch_idx, ckpt_save_path))
                if self.params['save_res'] and batch_idx > 0 and batch_idx % batches_per_save == 0:
                    self.save('{}/causal_pre_at_{}.{}'.format(self.save_dir, batch_idx, save_format), causal_pre)
        if self.params['save_res']:
            self.save('{}/causal_pre_final.{}'.format(self.save_dir,save_format), self.best_causal_pre)

        if self.params['binary_treatment']:
            self.ATE = np.mean(self.best_causal_pre)
            print('The average treatment effect (ATE) is', self.ATE)

    def evaluate(self, data, nb_intervals=200):
        """Internal evaluation in the training process of CausalEGM.

        Parameters
        ----------
        data
            List denoting the triplet data [X,Y,V] to be evaluated.
        nb_intervals
            Int object denoting number of intervals in continous treatment settings. Default: ``200``.

        Returns
        --------
        causal_pre
            Numpy.ndarray denoting the predicted individual treatment effect (ITE) or 
            values of average dose response function (ADRF).
        mse_x
            Float denoting treatment reconstruction loss.
        mse_y
            Float denoting outcome reconstruction loss.
        """ 
        data_x, data_y, data_v = data
        data_z_ = self.e_net.predict(data_v,verbose=0)
        data_z0 = data_z_[:,:self.params['z_dims'][0]]
        data_z1 = data_z_[:,self.params['z_dims'][0]:sum(self.params['z_dims'][:2])]
        data_z2 = data_z_[:,sum(self.params['z_dims'][:2]):sum(self.params['z_dims'][:3])]
        data_y_pred = self.f_net.predict(tf.concat([data_z0, data_z1, data_x], axis=-1),verbose=0)
        data_x_pred = self.h_net.predict(tf.concat([data_z0, data_z2], axis=-1),verbose=0)
        if self.params['binary_treatment']:
            data_x_pred = tf.sigmoid(data_x_pred)
        mse_x = np.mean((data_x-data_x_pred)**2)
        mse_y = np.mean((data_y-data_y_pred)**2)
        if self.params['binary_treatment']:
            #individual treatment effect (ITE) && average treatment effect (ATE)
            y_pred_pos = self.f_net.predict(tf.concat([data_z0, data_z1, np.ones((len(data_x),1))], axis=-1),verbose=0)
            y_pred_neg = self.f_net.predict(tf.concat([data_z0, data_z1, np.zeros((len(data_x),1))], axis=-1),verbose=0)
            ite_pre = y_pred_pos-y_pred_neg
            return ite_pre, mse_x, mse_y
        else:
            #average dose response function (ADRF)
            dose_response = []
            for x in np.linspace(self.params['x_min'], self.params['x_max'], nb_intervals):
                data_x = np.tile(x, (len(data_x), 1))
                y_pred = self.f_net.predict(tf.concat([data_z0, data_z1, data_x], axis=-1),verbose=0)
                dose_response.append(np.mean(y_pred))
            return np.array(dose_response), mse_x, mse_y
        
    def predict(self, data_x, data_v):
        """Predict the outcome given treatment and covariates in CausalEGM.

        Parameters
        ----------
        data_x
            Numpy.ndarray denoting treatment data with shape [nb_sample, 1] or [nb_sample, ].
        data_v
            Numpy.ndarray denoting covariants with shape [nb_sample, v_dim]. 

        Returns
        -------
        causal_pre
            Numpy.ndarray denoting the predicted potential outcome with shape [nb_sample, ].
        """ 
        assert len(data_x) == len(data_v)
        if len(data_x.shape)==1:
            data_x = data_x.reshape(-1,1)
        data_z_ = self.e_net.predict(data_v,verbose=0)
        data_z0 = data_z_[:,:self.params['z_dims'][0]]
        data_z1 = data_z_[:,self.params['z_dims'][0]:sum(self.params['z_dims'][:2])]
        data_y_pred = self.f_net.predict(tf.concat([data_z0, data_z1, data_x], axis=-1),verbose=0)
        return np.squeeze(data_y_pred)
    
    def getADRF(self, x_list, data_v=None):
        """Get average dosage response function (ADRF) in CausalEGM.

        Parameters
        ----------
        x_list
            List object denoting the treatment values.
        data_v
            Numpy.ndarray denoting covariants with shape [nb_sample, v_dim]. 
            If it is not provided, the covariants in the training data will be used.

        Returns
        -------
        causal_pre
            Numpy.ndarray denoting the predicted ADRF values with shape [nb_sample, ].
        """ 
        if data_v is None:
            data_v = self.data_sampler.load_all()[-1]
        data_z_ = self.e_net.predict(data_v,verbose=0)
        data_z0 = data_z_[:,:self.params['z_dims'][0]]
        data_z1 = data_z_[:,self.params['z_dims'][0]:sum(self.params['z_dims'][:2])]
        if not self.params['binary_treatment']:
            dose_response = []
            for x in x_list:
                data_x = np.tile(x, (len(data_v), 1))
                y_pred = self.f_net.predict(tf.concat([data_z0, data_z1, data_x], axis=-1),verbose=0)
                dose_response.append(np.mean(y_pred))
            return np.array(dose_response)
        else:
            print('ADRF is only applicable in continuous treatment setting!')
            sys.exit()
    
    def getCATE(self,data_v):
        """Get conditional average treatment effect (CATE) in CausalEGM.

        Parameters
        ----------
        data_v
            Numpy.ndarray denoting covariants with shape [nb_sample, v_dim]. 
            If it is not provided, the covariants in the training data will be used.

        Returns
        -------
        cate_pre
            Numpy.ndarray (1-D) denoting the predicted CATE values with shape [nb_sample, ].
        """ 
        assert data_v.shape[1] == self.params['v_dim']
        data_z_ = self.e_net.predict(data_v,verbose=0)
        data_z0 = data_z_[:,:self.params['z_dims'][0]]
        data_z1 = data_z_[:,self.params['z_dims'][0]:sum(self.params['z_dims'][:2])]
        data_z2 = data_z_[:,sum(self.params['z_dims'][:2]):sum(self.params['z_dims'][:3])]
        if self.params['binary_treatment']:
            y_pred_pos = self.f_net.predict(tf.concat([data_z0, data_z1, np.ones((len(data_v),1))], axis=-1),verbose=0)
            y_pred_neg = self.f_net.predict(tf.concat([data_z0, data_z1, np.zeros((len(data_v),1))], axis=-1),verbose=0)
            cate_pre = y_pred_pos-y_pred_neg
            return np.squeeze(cate_pre)
        else:
            print('CATE is only applicable in binary treatment setting!')
            sys.exit()
    

    def save(self, fname, data):
        """Save the data to the specified path."""

        if fname[-3:] == 'npy':
            np.save(fname, data)
        elif fname[-3:] == 'txt' or 'csv':
            np.savetxt(fname, data, fmt='%.6f')
        else:
            print('Wrong saving format, please specify either .npy, .txt, or .csv')
            sys.exit()

class VariationalCausalEGM(object):
    """Implementation of the variational CausalEGM model. 
    Instead of distribution match with GAN, we use variational inference in 
    the latent space.

    Parameters
    ----------
    params
        Dict object denoting the hyperparameters for deployments and building the model architecture. 
        See examples under the ``src/configs`` folder.
    timestamp
        Str object denoting the timestemp for specificing when the model is instanced. Default: ``None``.
    random_seed
        Int object denoting the random seed for controling randomness. Default: ``None``.

    Examples
    --------
    >>> from CausalEGM import VariationalCausalEGM, Sim_Hirano_Imbens_sampler
    >>> import yaml
    >>> params = yaml.safe_load(open('src/configs/Sim_Hirano_Imbens.yaml', 'r'))
    >>> x,y,v = Sim_Hirano_Imbens_sampler(batch_size=32).load_all()
    >>> model = VariationalCausalEGM(params=params,random_seed=12)
    >>> model.train(data=[x,y,v],n_iter=30000,save_format='npy')
    """
    def __init__(self, params, timestamp=None, random_seed=None):
        super(VariationalCausalEGM, self).__init__()
        self.params = params
        self.timestamp = timestamp
        if random_seed is not None:
            tf.keras.utils.set_random_seed(random_seed)
            os.environ['TF_DETERMINISTIC_OPS'] = '1'
        self.g_net = BaseFullyConnectedNet(input_dim=sum(params['z_dims']),output_dim = params['v_dim'], 
                                        model_name='g_net', nb_units=params['g_units'])
        self.e_net = BaseFullyConnectedNet(input_dim=params['v_dim'],output_dim = 2*sum(params['z_dims']), 
                                        model_name='e_net', nb_units=params['e_units'])

        self.f_net = BaseFullyConnectedNet(input_dim=1+params['z_dims'][0]+params['z_dims'][1],
                                        output_dim = 1, model_name='f_net', nb_units=params['f_units'])
        self.h_net = BaseFullyConnectedNet(input_dim=params['z_dims'][0]+params['z_dims'][2],
                                        output_dim = 1, model_name='h_net', nb_units=params['h_units'])

        self.g_e_optimizer = tf.keras.optimizers.Adam(params['lr'], beta_1=0.5, beta_2=0.9)
        self.d_optimizer = tf.keras.optimizers.Adam(params['lr'], beta_1=0.5, beta_2=0.9)
        self.z_sampler = Gaussian_sampler(mean=np.zeros(sum(params['z_dims'])), sd=1.0)

        self.initialize_nets()
        if self.timestamp is None:
            now = datetime.datetime.now(dateutil.tz.tzlocal())
            self.timestamp = now.strftime('%Y%m%d_%H%M%S')
        
        self.checkpoint_path = "{}/checkpoints/{}/{}".format(
            params['output_dir'], params['dataset'], self.timestamp)

        if self.params['save_model'] and not os.path.exists(self.checkpoint_path):
            os.makedirs(self.checkpoint_path)
        
        self.save_dir = "{}/results/{}/{}".format(
            params['output_dir'], params['dataset'], self.timestamp)

        if self.params['save_res'] and not os.path.exists(self.save_dir):
            os.makedirs(self.save_dir)

        self.ckpt = tf.train.Checkpoint(g_net = self.g_net,
                                   e_net = self.e_net,
                                   f_net = self.f_net,
                                   h_net = self.h_net,
                                   g_e_optimizer = self.g_e_optimizer,
                                   d_optimizer = self.d_optimizer)
        self.ckpt_manager = tf.train.CheckpointManager(self.ckpt, self.checkpoint_path, max_to_keep=3)                 

        if self.ckpt_manager.latest_checkpoint:
            self.ckpt.restore(self.ckpt_manager.latest_checkpoint)
            print ('Latest checkpoint restored!!') 
        
    def get_config(self):
        """Get the parameters CausalEGM model."""

        return {
                "params": self.params,
        }
    
    def initialize_nets(self, print_summary = False):
        """Initialize all the networks in CausalEGM."""

        self.g_net(np.zeros((1, sum(self.params['z_dims']))))
        self.e_net(np.zeros((1, self.params['v_dim'])))
        self.f_net(np.zeros((1, 1+self.params['z_dims'][0]+self.params['z_dims'][1])))
        self.h_net(np.zeros((1, self.params['z_dims'][0]+self.params['z_dims'][2])))
        if print_summary:
            print(self.g_net.summary())
            print(self.h_net.summary())
            print(self.f_net.summary())    
            print(self.h_net.summary()) 

    @tf.function
    def train_step(self, data_z, data_v, data_x, data_y):
        """Training step in the Variational CausalEGM model.

        Parameters
        ----------
        data_z
            Numpy.ndarray denoting latent features with shape [batch_size, z_dim]. 
        data_v
            Numpy.ndarray denoting covariants with shape [batch_size, v_dim]. 
        data_x
            Numpy.ndarray denoting treatment data with shape [batch_size, 1]. 
        data_y
            Numpy.ndarray denoting outcome data with shape [batch_size, 1]. 

        Returns
        --------
        logpv_z
            Float denoting likelihood for covariates.
        kl_loss
            Float denoting KL-divengence between p(z) and q(z|v).
        elbo
            Float denoting evidence lower bound (ELBO) loss.
        l2_loss_x
            Float denoting treatment reconstruction loss.
        l2_loss_y
            Float denoting outcome reconstruction loss.   
        g_e_loss
            Float denoting G E combined loss.      
        """ 
        with tf.GradientTape(persistent=True) as gen_tape:
            data_v_ = self.g_net(data_z)
            mean, logvar = self.encode(data_v)
            data_z_ = self.reparameterize(mean, logvar)
            data_v__ = self.g_net(data_z_)

            logpv_z = -tf.reduce_mean((data_v - data_v__)**2,axis=1)
            
            logqz_v = self.log_normal_pdf(data_z_, mean, logvar)
            logpz = self.log_normal_pdf(data_z_, 0., 0.)
            kl_loss = logqz_v-logpz # here it is not the formula of KL_loss, so will result in negative values
            elbo = tf.reduce_mean(logpv_z - kl_loss)
            
            data_z0 = data_z_[:,:self.params['z_dims'][0]]
            data_z1 = data_z_[:,self.params['z_dims'][0]:sum(self.params['z_dims'][:2])]
            data_z2 = data_z_[:,sum(self.params['z_dims'][:2]):sum(self.params['z_dims'][:3])]

            data_y_ = self.f_net(tf.concat([data_z0, data_z1, data_x], axis=-1))
            data_x_ = self.h_net(tf.concat([data_z0, data_z2], axis=-1))
            if self.params['binary_treatment']:
                data_x_ = tf.sigmoid(data_x_)
            l2_loss_x = tf.reduce_mean((data_x_ - data_x)**2)
            l2_loss_y = tf.reduce_mean((data_y_ - data_y)**2)
            g_e_loss = -elbo + self.params['beta']*(l2_loss_x+l2_loss_y)

        # Calculate the gradients for generators and discriminators
        g_e_gradients = gen_tape.gradient(g_e_loss, self.g_net.trainable_variables+self.e_net.trainable_variables+\
                                        self.f_net.trainable_variables+self.h_net.trainable_variables)
        
        # Apply the gradients to the optimizer
        self.g_e_optimizer.apply_gradients(zip(g_e_gradients, self.g_net.trainable_variables+self.e_net.trainable_variables+\
                                            self.f_net.trainable_variables+self.h_net.trainable_variables))
        return tf.reduce_mean(logpv_z), tf.reduce_mean(kl_loss), elbo, l2_loss_x, l2_loss_y, g_e_loss

    @tf.function
    def sample(self, eps=None):
        """Generate data by decoder."""
        if eps is None:
            eps = tf.random.normal(shape=(100, sum(self.params['z_dims'])))
        return self.g_net(eps)

    def encode(self, v):
        """Encode process and get both mean and variance."""
        mean, logvar = tf.split(self.e_net(v), num_or_size_splits=2, axis=1)
        return mean, logvar

    def reparameterize(self, mean, logvar):
        """Reparameterization for sample latent features."""
        eps = tf.random.normal(shape=mean.shape)
        return eps * tf.exp(logvar * .5) + mean
    
    def log_normal_pdf(self, sample, mean, logvar, axis=1):
        """Log likelihood of a normal distribution"""
        log2pi = tf.math.log(2. * np.pi)
        return tf.reduce_sum(
          -.5 * ((sample - mean) ** 2. * tf.exp(-logvar) + logvar + log2pi),
          axis=axis)
    
    def train(self, data=None, data_file=None, sep='\t', header=0, normalize=False,
            batch_size=32, n_iter=30000, batches_per_eval=500, batches_per_save=10000,
            startoff=0, verbose=1, save_format='txt'):
        """
        Train a variational CausalEGM model given the input data.
        
        Parameters
        ----------
        data
            List object containing the triplet data [X,Y,V]. Default: ``None``.
        data_file
            Str object denoting the path to the input file (csv, txt, npz).
        sep
            Str object denoting the delimiter for the input file. Default: ``\t``.
        header
            Int object denoting row number(s) to use as the column names. Default: ``0``.
        normalize
            Bool object denoting whether apply standard normalization to covariates. Default: ``False``.
        batch_size
            Int object denoting the batch size in training. Default: ``32``.
        n_iter
            Int object denoting the training iterations. Default: ``30000``.
        batches_per_eval
            Int object denoting the number of iterations per evaluation. Default: ``500``.
        batches_per_save
            Int object denoting the number of iterations per save. Default: ``10000``.
        startoff
            Int object denoting the beginning iterations to jump without save and evaluation. Defalt: ``0``.
        verbose
            Bool object denoting whether showing the progress bar. Default: ``False``.
        save_format
            Str object denoting the format (csv, txt, npz) to save the results. Default: ``txt``.
        
        """
        if data is None and data_file is None:
            self.data_sampler = Dataset_selector(self.params['dataset'])(batch_size=batch_size)
        elif data is not None:
            if len(data) != 3:
                print('Data imcomplete error, please provide pair-wise (X, Y, V) in a list or tuple.')
                sys.exit()
            else:
                self.data_sampler = Base_sampler(x=data[0],y=data[1],v=data[2],batch_size=batch_size,normalize=normalize)
        else:
            data = parse_file(data_file, sep, header, normalize)
            self.data_sampler = Base_sampler(x=data[0],y=data[1],v=data[2],batch_size=batch_size,normalize=normalize)

        best_loss = np.inf
        all_loss = []
        for batch_idx in range(n_iter+1):
            batch_x, batch_y, batch_v = self.data_sampler.next_batch()
            batch_z = self.z_sampler.get_batch(len(batch_x))
            #g_loss_adv, e_loss_adv, l2_loss_v, l2_loss_z, l2_loss_x, l2_loss_y, g_e_loss = self.train_gen_step(batch_z, batch_v, batch_x, batch_y)
            logpv_z, kl_loss, elbo, l2_loss_x, l2_loss_y, g_e_loss = self.train_step(batch_z, batch_v, batch_x, batch_y)
            all_loss.append([logpv_z, kl_loss, elbo, l2_loss_x, l2_loss_y, g_e_loss])
            if batch_idx % batches_per_eval == 0:
                loss_contents = '''Iteration [%d] : logpv_z [%.4f], kl_loss [%.4f],\
                elbo [%.4f], l2_loss_x [%.4f], l2_loss_y [%.4f], g_e_loss [%.4f]''' \
                %(batch_idx, logpv_z, kl_loss, elbo, l2_loss_x, l2_loss_y, g_e_loss)
                if verbose:
                    print(loss_contents)
                causal_pre, mse_x, mse_y = self.evaluate(self.data_sampler.load_all())
                if batch_idx >= startoff and mse_y < best_loss:
                    best_loss = mse_y
                    self.best_causal_pre = causal_pre
                    self.best_batch_idx = batch_idx
                    if self.params['save_model']:
                        ckpt_save_path = self.ckpt_manager.save(batch_idx)
                        #print('Saving checkpoint for iteration {} at {}'.format(batch_idx, ckpt_save_path))
                if self.params['save_res'] and batch_idx > 0 and batch_idx % batches_per_save == 0:
                    self.save('{}/causal_pre_at_{}.{}'.format(self.save_dir, batch_idx, save_format), causal_pre)
        if self.params['save_res']:
            self.save('{}/causal_pre_final.{}'.format(self.save_dir,save_format), self.best_causal_pre)
        if self.params['binary_treatment']:
            self.ATE = np.mean(self.best_causal_pre)
            print('The average treatment effect (ATE) is ', self.ATE)

    def evaluate(self, data, nb_intervals=200):
        """Internal evaluation in the training process of variational CausalEGM.

        Parameters
        ----------
        data
            List denoting the triplet data [X,Y,V] to be evaluated.
        nb_intervals
            Int object denoting number of intervals in continous treatment settings. Default: ``200``.

        Returns
        --------
        causal_pre
            Numpy.ndarray denoting the predicted individual treatment effect (ITE) or 
            values of average dose response function (ADRF).
        mse_x
            Float denoting treatment reconstruction loss.
        mse_y
            Float denoting outcome reconstruction loss.
        """ 
        data_x, data_y, data_v = data
        mean, logvar = self.encode(data_v)
        data_z_ = self.reparameterize(mean, logvar)
        data_z0 = data_z_[:,:self.params['z_dims'][0]]
        data_z1 = data_z_[:,self.params['z_dims'][0]:sum(self.params['z_dims'][:2])]
        data_z2 = data_z_[:,sum(self.params['z_dims'][:2]):sum(self.params['z_dims'][:3])]
        data_y_pred = self.f_net.predict(tf.concat([data_z0, data_z1, data_x], axis=-1),verbose=0)
        data_x_pred = self.h_net.predict(tf.concat([data_z0, data_z2], axis=-1),verbose=0)
        if self.params['binary_treatment']:
            data_x_pred = tf.sigmoid(data_x_pred)
        mse_x = np.mean((data_x-data_x_pred)**2)
        mse_y = np.mean((data_y-data_y_pred)**2)
        if self.params['binary_treatment']:
            #individual treatment effect (ITE) && average treatment effect (ATE)
            y_pred_pos = self.f_net.predict(tf.concat([data_z0, data_z1, np.ones((len(data_x),1))], axis=-1),verbose=0)
            y_pred_neg = self.f_net.predict(tf.concat([data_z0, data_z1, np.zeros((len(data_x),1))], axis=-1),verbose=0)
            ite_pre = y_pred_pos-y_pred_neg
            return ite_pre, mse_x, mse_y
        else:
            #average dose response function (ADRF)
            dose_response = []
            for x in np.linspace(self.params['x_min'], self.params['x_max'], nb_intervals):
                data_x = np.tile(x, (len(data_x), 1))
                y_pred = self.f_net.predict(tf.concat([data_z0, data_z1, data_x], axis=-1),verbose=0)
                dose_response.append(np.mean(y_pred))
            return np.array(dose_response), mse_x, mse_y

    def save(self, fname, data):
        """Save the data to the specified path."""
        if fname[-3:] == 'npy':
            np.save(fname, data)
        elif fname[-3:] == 'txt' or 'csv':
            np.savetxt(fname, data, fmt='%.6f')
        else:
            print('Wrong saving format, please specify either .npy, .txt, or .csv')
            sys.exit()


import numpy as np
import math
import os
import sys
import pandas as pd
from sklearn.preprocessing import MinMaxScaler,MaxAbsScaler,StandardScaler
from sklearn.model_selection import train_test_split
from scipy.sparse import diags
from scipy.stats import norm

def Dataset_selector(name):
    if name == 'Semi_acic':
        return Semi_acic_sampler
    elif name=='Semi_ihdp':
        return Semi_ihdp_sampler
    elif name=='Sim_Hirano_Imbens':
        return Sim_Hirano_Imbens_sampler
    elif name=='Sim_Sun':
        return Sim_Sun_sampler
    elif name=='Sim_Colangelo':
        return Sim_Colangelo_sampler
    elif name=='Semi_Twins':
        return Semi_Twins_sampler
    else:
        print('Cannot find the example data sampler: %s!'%name)
        sys.exit()

class Base_sampler(object):
    """Base data sampler.

    Parameters
    ----------
    x
        List or Numpy.ndarray bject denoting the treatment with length N or shape (N, 1) or (N, ). 
    y
        List or Numpy.ndarray bject denoting the outcome with length N or shape (N, 1) or (N, ). 
    v
        List or Numpy.ndarray bject denoting the covariates with length N or shape (N, v_dim). 
    batch_size
        Int object denoting the batch size for mini-batch training. Default: ``32``.

    Examples
    --------
    >>> from CausalEGM import Base_sampler
    >>> import numpy as np
    >>> x = np.random.normal(size=(2000,))
    >>> y = np.random.normal(size=(2000,))
    >>> v = np.random.normal(size=(2000,100))
    >>> ds = Base_sampler(x=x,y=y,v=v)
    >>> batch = ds.next_batch() # get a batch of data
    >>> data = ds.load_all() # get all data as a triplet
    """
    def __init__(self, x, y, v, batch_size=32, normalize=False, random_seed=123):
        assert len(x)==len(y)==len(v)
        np.random.seed(random_seed)
        self.data_x = np.array(x, dtype='float32')
        self.data_y = np.array(y, dtype='float32')
        self.data_v = np.array(v, dtype='float32')
        if len(self.data_x.shape) == 1:
            self.data_x = self.data_x.reshape(-1,1)
        if len(self.data_y.shape) == 1:
            self.data_y = self.data_y.reshape(-1,1)
        self.batch_size = batch_size
        if normalize:
            self.data_v = StandardScaler().fit_transform(self.data_v)
        self.sample_size = len(x)
        self.full_index = np.arange(self.sample_size)
        np.random.shuffle(self.full_index)
        self.idx_gen = self.create_idx_generator(sample_size=self.sample_size)
        
    def create_idx_generator(self, sample_size, random_seed=123):
        while True:
            for step in range(math.ceil(sample_size/self.batch_size)):
                if (step+1)*self.batch_size <= sample_size:
                    yield self.full_index[step*self.batch_size:(step+1)*self.batch_size]
                else:
                    yield np.hstack([self.full_index[step*self.batch_size:],
                                    self.full_index[:((step+1)*self.batch_size-sample_size)]])
                    np.random.shuffle(self.full_index)

    def next_batch(self):
        indx = next(self.idx_gen)
        return self.data_x[indx,:], self.data_y[indx,:], self.data_v[indx, :]
    
    def load_all(self):
        return self.data_x, self.data_y, self.data_v

class Semi_acic_sampler(Base_sampler):
    """ACIC 2018 competition dataset (binary treatment) sampler (inherited from Base_sampler).

    Parameters
    ----------
    batch_size
        Int object denoting the batch size for mini-batch training. Default: ``32``.
    path
        Str object denoting the path to the original dataset.
    ufid
        Str object denoting the unique id of a specific semi-synthetic setting.
    Examples
    --------
    >>> from CausalEGM import Semi_acic_sampler
    >>> import numpy as np
    >>> x = np.random.normal(size=(2000,))
    >>> y = np.random.normal(size=(2000,))
    >>> v = np.random.normal(size=(2000,100))
    >>> ds = Semi_acic_sampler(path='../data/ACIC_2018',ufid='d5bd8e4814904c58a79d7cdcd7c2a1bb')
    """
    def __init__(self, batch_size=32, path='../data/ACIC_2018', 
                ufid='d5bd8e4814904c58a79d7cdcd7c2a1bb'):
        self.df_covariants = pd.read_csv('%s/x.csv'%path, index_col='sample_id',header=0, sep=',')
        self.df_sim = pd.read_csv('%s/scaling/factuals/%s.csv'%(path, ufid),index_col='sample_id',header=0, sep=',')
        dataset = self.df_covariants.join(self.df_sim, how='inner')
        x = dataset['z'].values.reshape(-1,1)
        y = dataset['y'].values.reshape(-1,1)
        v = dataset.values[:,:-2]
        super().__init__(x,y,v,batch_size=batch_size,normalize=True)

class Sim_Hirano_Imbens_sampler(Base_sampler):
    """Hirano Imbens simulation dataset (continuous treatment) sampler (inherited from Base_sampler).

    Parameters
    ----------
    batch_size
        Int object denoting the batch size for mini-batch training. Default: ``32``.
    N
        Sample size. Default: ``20000``.
    v_dim
        Int object denoting the dimension for covariates. Default: ``200``.
    seed
        Int object denoting the random seed. Default: ``0``.
    Examples
    --------
    >>> from CausalEGM import Sim_Hirano_Imbens_sampler
    >>> ds = Sim_Hirano_Imbens_sampler(batch_size=32, N=20000, v_dim=200, seed=0)
    """
    def __init__(self, batch_size=32, N=20000, v_dim=200, seed=0):
        np.random.seed(seed)
        v = np.random.exponential(scale=1.0, size=(N, v_dim))
        rate = v[:,0] + v[:,1]
        scale = 1/rate
        x = np.random.exponential(scale=scale)
        y = np.random.normal(x + (v[:,0] + v[:,2]) * np.exp(-x * (v[:,0] + v[:,2])) , 1)
        x = x.reshape(-1,1)
        y = y.reshape(-1,1)
        super().__init__(x,y,v,batch_size=batch_size,normalize=True)

class Sim_Sun_sampler(Base_sampler):
    """Sun simulation dataset (continuous treatment) sampler (inherited from Base_sampler).

    Parameters
    ----------
    batch_size
        Int object denoting the batch size for mini-batch training. Default: ``32``.
    N
        Sample size. Default: ``20000``.
    v_dim
        Int object denoting the dimension for covariates. Default: ``200``.
    seed
        Int object denoting the random seed. Default: ``0``.
    Examples
    --------
    >>> from CausalEGM import Sim_Sun_sampler
    >>> ds = Sim_Sun_sampler(batch_size=32, N=20000, v_dim=200, seed=0)
    """
    def __init__(self, batch_size, N=20000, v_dim=200, seed=0):
        np.random.seed(seed)
        v = np.random.normal(0, 1, size=(N, v_dim))        
        x = np.random.normal(-2*(np.sin(2*v[:,0]))+ ((v[:,1])**2 - 1/3) + (v[:,2]-1/2) + np.cos(v[:,3]), 1)
        y = np.random.normal(((v[:,0] - 1/2)+ np.cos(v[:,1]) + (v[:,4])**2 + (v[:,5])) + x, 1)       
        x = x.reshape(-1,1)
        y = y.reshape(-1,1)
        super().__init__(x,y,v,batch_size=batch_size,normalize=True)

class Sim_Colangelo_sampler(Base_sampler):
    """Colangelo simulation dataset (continuous treatment) sampler (inherited from Base_sampler).

    Parameters
    ----------
    batch_size
        Int object denoting the batch size for mini-batch training. Default: ``32``.
    N
        Sample size. Default: ``20000``.
    v_dim
        Int object denoting the dimension for covariates. Default: ``200``.
    seed
        Int object denoting the random seed. Default: ``0``.
    Examples
    --------
    >>> from CausalEGM import Sim_Colangelo_sampler
    >>> ds = Sim_Colangelo_sampler(batch_size=32, N=20000, v_dim=100, seed=0)
    """
    def __init__(self, batch_size=32, N=20000, v_dim=100, seed=0,
                rho=0.5, offset = [-1,0,1], d=1, a=3, b=0.75):
        np.random.seed(seed)
        k = np.array([rho*np.ones(v_dim-1),np.ones(v_dim),rho*np.ones(v_dim-1)])
        sigma = diags(k,offset).toarray()
        theta = np.array([(1/(l**2)) for l in list(range(1,(v_dim+1)))])
        epsilon = np.random.normal(0,1,N)
        nu = np.random.normal(0,1,N)
        v = np.random.multivariate_normal(np.zeros(v_dim),sigma,size=[N,])
        x = d*norm.cdf((a*v@theta)) + b*nu - 0.5
        y = 1.2*x + (x**3) + (x*v[:,0]) + 1.2*(v@theta) + epsilon
        x = x.reshape(-1,1)
        y = y.reshape(-1,1)
        super().__init__(x,y,v,batch_size=batch_size,normalize=True)

class Semi_Twins_sampler(Base_sampler):
    """Twins semi synthetic  dataset sampler (inherited from Base_sampler).

    Parameters
    ----------
    batch_size
        Int object denoting the batch size for mini-batch training. Default: ``32``.
    seed
        Int object denoting the random seed. Default: ``0``.
    path
        Str obejct denoting the path to the original data.
    Examples
    --------
    >>> from CausalEGM import Semi_Twins_sampler
    >>> ds = Semi_Twins_sampler(batch_size=32, path='../data/Twins')
    """
    def __init__(self, batch_size=32, seed=0,
                path='../data/Twins'):
        covariate_df = pd.read_csv('%s/twin_pairs_X_3years_samesex.csv'%path).iloc[:,2:].drop(['infant_id_0', 'infant_id_1'], axis=1)
        treatment_df_ = pd.read_csv('%s/twin_pairs_T_3years_samesex.csv'%path).iloc[:,1:]
        outcome_df = pd.read_csv('%s/twin_pairs_Y_3years_samesex.csv'%path).iloc[:,1:]
        #### discard NAN values
        rows_with_nan = [index for index, row in covariate_df.iterrows() if row.isnull().any()]
        covariate_df = covariate_df.drop(rows_with_nan)
        treatment_df_ = treatment_df_.drop(rows_with_nan)
        outcome_df = outcome_df.drop(rows_with_nan)
        #### select those below 2kg:
        rows_less2kg = [index for index, row in treatment_df_.iterrows() if (row['dbirwt_1']>=2000)]
        covariate_df = covariate_df.drop(rows_less2kg)
        treatment_df_ = treatment_df_.drop(rows_less2kg)
        outcome_df = outcome_df.drop(rows_less2kg)

        x = np.concatenate([treatment_df_.values[:,0], treatment_df_.values[:,1]])/1000
        v =  np.concatenate([covariate_df.values, covariate_df.values])
        np.random.seed(seed)
        eps = np.random.normal(0, 0.25, size=(v.shape[0],))
        gamma = np.random.normal(0, 0.025, size=(v.shape[1],))
        y = -2 * 1/(1 + np.exp(-3 * x)) + np.dot(v, gamma) + eps
        auxiliary_constant =  np.mean(np.dot(v, gamma))
        x = x.reshape(-1,1)
        y = y.reshape(-1,1)
        super().__init__(x,y,v,batch_size=batch_size,normalize=True)

class Gaussian_sampler(object):
    def __init__(self, mean, sd=1, N=20000):
        self.total_size = N
        self.mean = mean
        self.sd = sd
        np.random.seed(1024)
        self.X = np.random.normal(self.mean, self.sd, (self.total_size,len(self.mean)))
        self.X = self.X.astype('float32')
        self.Y = None

    def train(self, batch_size, label = False):
        indx = np.random.randint(low = 0, high = self.total_size, size = batch_size)
        return self.X[indx, :]

    def get_batch(self,batch_size):
        return np.random.normal(self.mean, self.sd, (batch_size,len(self.mean))).astype('float32')

    def load_all(self):
        return self.X, self.Y

class Assump_valid_sampler(object):
    def __init__(self, v_dim, z_dim, N=100000, n_heldout=10000, random_seed=123):
        from scipy.stats import ortho_group
        np.random.seed(random_seed)
        self.sample_size = N
        #preset diagonal matrix M
        eigenvalues = np.hstack([np.linspace(5,4,10),np.linspace(0.1,0.01,v_dim-10)])
        M = np.diag(eigenvalues)
        PCA_recon = sum(eigenvalues[z_dim:])
        print('PCA_reconstruction error: %.5f'%PCA_recon)
        #randomly generating othornomal bases and generate V
        U = ortho_group.rvs(v_dim)
        self.Sigma = np.dot(np.dot(U,M), U.T)
        self.mu = np.random.uniform(low=-1.0, high=1.0,size=(v_dim,))
        V = np.random.multivariate_normal(mean=self.mu, cov=self.Sigma,size = self.sample_size)
        V_heldout = np.random.multivariate_normal(mean=self.mu, cov=self.Sigma,size = n_heldout)
        #construct features of V
        self.A = np.dot(np.diag(eigenvalues**(-0.5)),U.T)
        T = np.dot(self.A, (V-self.mu).T).T
        T_heldout = np.dot(self.A, (V_heldout-self.mu).T).T
        self.A = self.A.astype('float32')
        self.mu = self.mu.astype('float32')
        self.data_v = V.astype('float32')
        self.data_t = T.astype('float32')
        self.data_v_heldout = V_heldout.astype('float32')
        self.data_t_heldout = T_heldout.astype('float32')

    def train(self, batch_size):
        indx = np.random.randint(low = 0, high = self.sample_size, size = batch_size)
        return self.data_v[indx, :], self.data_t[indx, :]
    def load_all(self):
        return self.data_v, self.data_t


def parse_file(path, sep='\t', header = 0, normalize=True):
    assert os.path.exists(path)
    if path[-3:] == 'npz':
        data = np.load(path)
        data_x, data_y, data_v = data['x'],data['y'],data['v']
    elif  path[-3:] == 'csv':
        data = pd.read_csv(path, header=0, sep=sep).values
        data_x = data[:,0].reshape(-1, 1).astype('float32')
        data_y = data[:,1].reshape(-1, 1).astype('float32')
        data_v = data[:,2:].astype('float32')
    elif path[-3:] == 'txt':
        data = np.loadtxt(path,delimiter=sep)
        data_x = data[:,0].reshape(-1, 1).astype('float32')
        data_y = data[:,1].reshape(-1, 1).astype('float32')
        data_v = data[:,2:].astype('float32')
    else:
        print('File format not recognized, please use .npz, .csv or .txt as input.')
        sys.exit()
    if normalize:
        data_v = StandardScaler().fit_transform(data_v)
    return data_x, data_y, data_v



import setuptools

setuptools.setup(
    name="CausalEGM", 
    version="0.4.0",
    author="Qiao Liu",
    author_email="liuqiao@stanford.edu",
    description="CausalEGM: an encoding generative modeling approach to dimension reduction and covariate adjustment in causal inference with observational studies",
    long_description="In this article, we develop CausalEGM, a deep learning framework for nonlinear dimension reduction and generative modeling of the dependency among covariate features affecting treatment and response. CausalEGM can be used for estimating causal effects in both binary and continuous treatment settings. By learning a bidirectional transformation between the high-dimensional covariate space and a low-dimensional latent space and then modeling the dependencies of different subsets of the latent variables on the treatment and response, CausalEGM can extract the latent covariate features that affect both treatment and response. By conditioning on these features, one can mitigate the confounding effect of the high dimensional covariate on the estimation of the causal relation between treatment and response. In a series of experiments, the proposed method is shown to achieve superior performance over existing methods in both binary and continuous treatment settings. The improvement is substantial when the sample size is large and the covariate is of high dimension. Finally, we established excess risk bounds and consistency results for our method, and discuss how our approach is related to and improves upon other dimension reduction approaches in causal inference. CausalEGM is freely available at https://github.com/SUwonglab/CausalEGM.",
    long_description_content_type="text/markdown",
    url="https://github.com/SUwonglab/CausalEGM",
    packages=setuptools.find_packages(),
    install_requires=[
   'tensorflow>=2.8.0',
   'scikit-learn',
   'pandas',
   'python-dateutil'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.9',
    entry_points={
    'console_scripts': [
        'causalEGM = CausalEGM.cli:main',
    ]},
)
